# Fern
# 
# @Author: Lin, Max
# @Email : max_lin1@dell.com
# @Time  : 2021/9/26 10:40 上午
#
# =============================================================================
"""applications.py"""
import tensorflow_text      # for hub.load
import tensorflow as tf
import tensorflow_hub as hub

from fern.logging import Logging
from fern.utils.tensor_ops import map_flat_values


logger = Logging()


class BertEncoder(tf.keras.layers.Layer):
    """BERT语言预训练模型, 默认使用中文"""
    _preprocessor = None
    _encoder = None
    model_config = {
        'bert_en_uncased_L-12_H-768_A-12': {
            'preprocessor_uri': 'https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/3',
            'encoder_uri': 'https://tfhub.dev/tensorflow/bert_en_uncased_L-12_H-768_A-12/4'
        },
        'bert_en_uncased_L-24_H-1024_A-16': {
            'preprocessor_uri': 'https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/3',
            'encoder_uri': 'https://tfhub.dev/tensorflow/bert_en_uncased_L-24_H-1024_A-16/4'
        },
        'bert_en_wwm_uncased_L-24_H-1024_A-16': {
            'preprocessor_uri': 'https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/3',
            'encoder_uri': 'https://tfhub.dev/tensorflow/bert_en_wwm_uncased_L-24_H-1024_A-16/4'
        },
        'bert_en_cased_L-12_H-768_A-12': {
            'preprocessor_uri': 'https://tfhub.dev/tensorflow/bert_en_cased_preprocess/3',
            'encoder_uri': 'https://tfhub.dev/tensorflow/bert_en_cased_L-12_H-768_A-12/4'
        },
        'bert_en_cased_L-24_H-1024_A-16': {
            'preprocessor_uri': 'https://tfhub.dev/tensorflow/bert_en_cased_preprocess/3',
            'encoder_uri': 'https://tfhub.dev/tensorflow/bert_en_cased_L-24_H-1024_A-16/4'
        },
        'bert_en_wwm_cased_L-24_H-1024_A-16': {
            'preprocessor_uri': 'https://tfhub.dev/tensorflow/bert_en_cased_preprocess/3',
            'encoder_uri': 'https://tfhub.dev/tensorflow/bert_en_wwm_cased_L-24_H-1024_A-16/4'
        },
        'bert_zh_L-12_H-768_A-12': {
            'preprocessor_uri': 'https://tfhub.dev/tensorflow/bert_zh_preprocess/3',
            'encoder_uri': 'https://tfhub.dev/tensorflow/bert_zh_L-12_H-768_A-12/4'
        },
        'bert_multi_cased_L-12_H-768_A-12': {
            'preprocessor_uri': 'https://tfhub.dev/tensorflow/bert_multi_cased_preprocess/3',
            'encoder_uri': 'https://tfhub.dev/tensorflow/bert_multi_cased_L-12_H-768_A-12/4'
        }
    }

    def __init__(self, seq_len: int, model_name: str, **kwargs):
        """
        根据配置生成Bert模型层

        Args:
            seq_len: 需要配置的最大输入长度, 默认128, 范围是1-512
            model_name: 在`BertEncoder.model_config.keys()`中选取
            **kwargs: keras layer通用配置
        """
        super().__init__(**kwargs)
        self.seq_len = seq_len
        self.model_name = model_name
        self.set_model(model_name)

        self.tokenize = self._preprocessor.tokenize
        self.bert_input_layer = hub.KerasLayer(
            self._preprocessor.bert_pack_inputs,
            arguments=dict(seq_length=seq_len),
            name='bert-input-layer')  # Optional argument.

    def get_config(self):
        config = super().get_config()
        config.update({
            'seq_len': self.seq_len,
            'model_name': self.model_name})
        return config

    def set_model(self, name):
        """
        修改模型和模型对应的preprocessor方法

        Args:
            name: 根据BertEncoder.uri_config.keys()获取合适的配置, 把key输入就可以完成模型的初始化
        """
        if name not in self.model_config:
            raise KeyError(f'无法找到合适的key: {self.uri_config.keys()}')
        preprocessor_uri = self.model_config[name]['preprocessor_uri']
        encoder_uri = self.model_config[name]['encoder_uri']
        self._preprocessor = hub.load(preprocessor_uri)
        self._encoder = hub.load(encoder_uri)

    def call(self, inputs, **kwargs):
        """
        批量文本转化为token

        Args:
            inputs: 支持输入任意shape的array, 支持ragged tensor; 例如 `[['text 1', 'text 2'], ['text 3']]`
            **kwargs:

        Returns:
            输入tensor则返回tensor, 输入ragged tensor则返回ragged tensor
        """
        encode_output = map_flat_values(self.map_func, inputs)
        return encode_output

    def map_func(self, task_logs):
        """
        输入string list, 输出bert转化后的输出

        Args:
            task_logs: shape [all_task_size, ], 所有当前batch的所有task log列表

        Returns:
            [all_task_size, encoder_length]
        """
        txt_tokenized = self.tokenize(task_logs)
        bert_input = self.bert_input_layer([txt_tokenized])
        bert_output = self._encoder(bert_input)['sequence_output']  # [task_size, seq_len, 768]
        res = bert_output[:, 0, :]  # [task_size, 768]
        return res



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""data classes"""
from typing import *
import csv
from multiprocessing.pool import Pool

import pandas as pd
from fern.utils import check_path


class FernSeries(pd.Series):
    @property
    def _constructor(self):
        return FernSeries

    @property
    def _constructor_expanddim(self):
        return FernDataFrame

    def parallel_map(self, func: Callable, args: Optional[Tuple] = None, processes: Optional[int] = None):
        """
        map method that use multiple processes

        Args:
            func: called to preprocess data
            args: args to be fed into the func, in addition to the original data
            processes: how many cpu to be used, default use all cpu

        Returns:
            processed series

        Examples:
            >>> d = FernSeries([1,2,3,4])
            >>> def multiply(x, y):
            ...     return x * y
            >>> d.parallel_map(func=multiply, args=(2, ))
            0    2
            1    4
            2    6
            3    8
            dtype: int64
        """
        args = args if args is not None else tuple()
        with Pool(processes) as pool:
            multi_res = [pool.apply_async(func, (item, *args)) for item in self.array]
            new_array = [data.get() for data in multi_res]
        return FernSeries(data=new_array, index=self.index)


class FernDataFrame(pd.DataFrame):
    """sub class of pandas.DataFrame with additional function"""

    @property
    def _constructor_expanddim(self):
        """for lint check"""
        raise NotImplementedError

    @property
    def _constructor(self):
        return FernDataFrame

    @property
    def _constructor_sliced(self):
        return FernSeries

    def parallel_apply(
        self,
        func: Callable[[Dict], Any],
        args: Optional[Tuple] = None,
        axis: int = 1,
        processes: Optional[int] = None
    ) -> FernSeries:
        """
        apply method which uses multiple processes

        Args:
            func: Function to apply to each column or row.
            args: args to be fed into the func, in addition to the original data
            axis: Axis along which the function is applied
                * 0: apply function to each column.
                * 1: apply function to each row.
            processes: how many cpu to be used, default use all cpu

        Returns:
            processed series
        """
        data: FernSeries = self.apply(lambda x: x.to_dict(), axis=axis)
        res = data.parallel_map(func=func, args=args, processes=processes)
        return res

    def save(self, path):
        """
        save data frame with csv format to path

        Parameters
        ----------
        path : str, Path
            path where data save
        """
        check_path(path)
        self.to_csv(path, index=True, encoding='utf-8-sig', quoting=csv.QUOTE_NONNUMERIC)



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""logging tool"""
from typing import *
import logging
from pathlib import Path


class Logging(object):
    """
    配置日志类, 如果是父类 `'.' not in name`, 那么允许继承, 不允许继承

    由于经常有程序修改 root logger 的配置(如tensorflow), 所有这里会默认屏蔽所有root logger的所有配置

    Examples
    ----
    - 在项目的根目录实例化logger::

        >>> from fern.logging import Logging
        >>> logger = Logging(__name__)
        >>> logger.add_stream_handler()

    - 在任何需要引用使用日志的地方执行::

        >>> import logging    # 注意是logging, 第一步已经把需要的信息注册到logging系统了, 可以无缝使用
        >>> logger = logging.getLogger(__name__)
    """

    formatter = logging.Formatter(fmt='%(asctime)s - %(name)s: %(message)s')
    levels = dict(
        DEBUG=logging.DEBUG,
        INFO=logging.INFO,
        WARN=logging.WARN,
        WARNING=logging.WARN,
        ERROR=logging.ERROR
    )

    def __init__(self, name: str = 'Fern', level: Union[str, int] = 'info', propagate=False):
        logger = logging.getLogger(name)

        # 所有logger都不从root中继承handler, 但如果是子logger则要从父logger中继承
        if isinstance(name, str):
            logger.propagate = ('.' in name) or propagate
        # set level
        if isinstance(level, int):
            logger.setLevel(level)
        else:
            logger.setLevel(self.levels.get(str(level).upper(), 'WARN'))
        # add logger
        self.__logger = logger
        self.add_stream_handler()

    def __getattr__(self, item: str):
        if item in dir(self):
            # dir取出实例的所有属性
            return getattr(self, item)
        else:
            # 如果这个属性不在这个类里面, 那么就是__logger里面找, 用户logger.info的方案
            return getattr(self.__logger, item)

    def __repr__(self):
        return repr(self.__logger)

    def add_stream_handler(self, level: Union[str, int, None] = None, **kwargs):
        """
        日志导出到终端中, 新的stream handler会覆盖旧的

        Args:
            level: 默认的level是比 DEBUG 还小的 0
            **kwargs: other file handle config
        """
        stream_handler = logging.StreamHandler(**kwargs)
        stream_handler.setFormatter(fmt=self.formatter)

        if level in self.levels:
            level = self.levels[level]
        elif isinstance(level, int):
            level = level
        else:
            level = 0

        stream_handler.setLevel(level)

        # 避免重复添加handlers
        handlers = [handler for handler in self.__logger.handlers if not isinstance(handler, logging.StreamHandler)]
        handlers.append(stream_handler)
        self.__logger.handlers = handlers

    def add_file_handle(self, filename: Union[str, Path] = 'base.log', level: Union[str, int, None] = None, **kwargs):
        """
        日志导出到文件系统中, 新的file handler会覆盖旧的

        Args:
            filename: 文件路径
            level: 默认的level是比 DEBUG 还小的 0
            **kwargs: other file handle config
        """
        if not Path(filename).parent.exists():
            Path(filename).parent.mkdir(parents=True)

        file_handler = logging.FileHandler(filename=filename, **kwargs)
        file_handler.setFormatter(self.formatter)

        if level in self.levels:
            level = self.levels[level]
        elif isinstance(level, int):
            level = level
        else:
            level = 0

        file_handler.setLevel(level)

        # 避免重复添加handlers
        handlers = [handler for handler in self.__logger.handlers if not isinstance(handler, logging.FileHandler)]
        handlers.append(file_handler)
        self.__logger.handlers = handlers


class LoggingContext(object):
    """
    通过通过上下文, 执行前和执行结束打印结果

    Examples
    ----
    >>> with LoggingContext(print, 'test'):
    ...     print('xxx')
    Begin: test
    xxx
    End: test
    """
    def __init__(self, log_func: Callable, msg: str):
        """
        Args:
            log_func: logging.info等函数
            msg: 需要打印的数据
        """
        self.log_func = log_func
        self.msg = msg

    def __enter__(self):
        self.log_func(f'Begin: {self.msg}')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.log_func(f'End: {self.msg}')



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""data balance"""
from typing import *
from copy import deepcopy

import numpy as np
import pandas as pd


def over_sample(data: pd.DataFrame, label_col: str, rate: float, random_state: Optional[int] = None) -> pd.DataFrame:
    """
    过采样以最大值为基准：
        1. 样本数低于比例的都会被over sample
        2. 样本数高于于比例的不变

    支持的标签类型：
        1. 多任务: {'task1': array([1, 1, 0], 'task2': array([1, 0, 0]}
        2. 单任务多标签: array([1, 1, 0])
        3. 单任务单标签: array([1, 0, 0])

    Args:
        data: 需要上采样的数据
        label_col: 需要处理的标签列名
        rate: over sample之后，所有的类别的数量不小于 rate x 最大类别样本数
        random_state: 随机种子

    Returns:
        采样之后的列数据
    """
    res_data = deepcopy(data)
    label_example = data[label_col].iloc[0]
    if isinstance(label_example, dict):
        key = list(label_example.keys())[0]

    def generate_label_evaluate_standard(item):
        if isinstance(item, dict):
            item = item[key]
        return item

    labels = np.vstack(data[label_col].map(generate_label_evaluate_standard))
    count_label = np.sum(labels, axis=0)
    target_num = int(max(count_label) * rate)
    # 遍历所有的样本数量，并随机采样到至少target num
    for col_idx, num in sorted(enumerate(count_label), key=lambda item: item[1]):
        row_idx = np.argwhere(labels[:, col_idx] == 1).reshape(-1)
        row_idx = np.random.RandomState(random_state).permutation(row_idx)
        if num >= target_num or num < 2:
            continue
        repeat_times = int(target_num // num - 1)
        copy_num = int(target_num % num)

        _data = data.iloc[row_idx]
        res_data.append(_data.loc[_data.index.repeat(repeat_times)])
        res_data.append(_data.iloc[:copy_num])

    random_idx = np.random.RandomState(random_state).permutation(res_data.shape[0])
    res_data = res_data.iloc[random_idx]
    return res_data



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""model trainer"""
import logging
import pickle
from pathlib import Path
from typing import *

import numpy as np
import tensorflow as tf
from tensorflow.keras.metrics import Metric

from fern.models.model import FernModel
from fern.utils.common import ProgressBar


logger = logging.getLogger('Fern')


class FernBaseTrainer(object):
    model: FernModel
    optimizer: Optional[tf.optimizers.Optimizer]
    data: Dict[str, Union[tf.data.Dataset, int]]
    metrics: Dict[str, Dict[str, Metric]]
    early_stop_metric: Metric
    label_weight: Union[Dict[str, np.ndarray], np.ndarray]

    def train(self, epochs: int = 1, early_stop: Optional[int] = None, mode: str = 'search') -> Tuple[float, int]:
        """
        train the model

        Args:
            epochs: epoch number
            early_stop:
                If None, no early stop will be used.
                If is number, after number times no update, will stop training early.
            mode:
                search: train model with data_train
                server: train model with data_train and data_val

        Returns:
            (best_score, best_epoch)
        """
        raise NotImplementedError

    @staticmethod
    def load_data(
            path: Union[str, Path],
            batch_size: int,
            data_col: str,
            label_col: str,
    ) -> Dict[str, Union[tf.data.Dataset, int]]:
        """
        load dataset from path

        if there are multiple labels or data, for example data={'col1': data1, 'col2': data2}
        and label={'col3': label1, 'col4': label2}, then the output of the data set will be dict too, which looks like
        data_batch={'col1': data1, 'col2': data2} and label_batch={'col3': label1, 'col4': label2}

        Args:
            path: Points to training data in npz format
            batch_size: batch size
            data_col: data column name
            label_col: label column name

        Returns
        -------
        dict[str, tf.data.Dataset|int]
            a dictionary with dataset_train, dataset_val, dataset_total, step_train, step_val and step_total
        """
        with open(path, 'rb') as f:
            data = pickle.load(f)
            data_total: Union[Dict[str, np.ndarray], np.ndarray] = data[f'{data_col}_total']
            label_total: Union[Dict[str, np.ndarray], np.ndarray] = data[f'{label_col}_total']

            data_train: Union[Dict[str, np.ndarray], np.ndarray] = data[f'{data_col}_train']
            label_train: Union[Dict[str, np.ndarray], np.ndarray] = data[f'{label_col}_train']

            data_val: Union[Dict[str, np.ndarray], np.ndarray] = data[f'{data_col}_val']
            label_val: Union[Dict[str, np.ndarray], np.ndarray] = data[f'{label_col}_val']

        if isinstance(data_total, dict):
            # multi input
            len_total = list(data_total.values())[0].shape[0]
            len_train = list(data_train.values())[0].shape[0]
            len_val = list(data_val.values())[0].shape[0]
        else:
            # single input
            len_total = data_total.shape[0]
            len_train = data_train.shape[0]
            len_val = data_val.shape[0]

        dataset_train = tf.data.Dataset.from_tensor_slices((data_train, label_train)).shuffle(len_train)\
            .batch(batch_size)
        dataset_val = tf.data.Dataset.from_tensor_slices((data_val, label_val)).batch(batch_size)
        dataset_total = tf.data.Dataset.from_tensor_slices((data_total, label_total)).shuffle(len_total)\
            .batch(batch_size)

        step_train = int(np.ceil(len_train / batch_size))
        step_val = int(np.ceil(len_val / batch_size))
        step_total = int(np.ceil(len_total / batch_size))

        data = {
            'dataset_train': dataset_train,
            'dataset_val': dataset_val,
            'dataset_total': dataset_total,
            'step_train': step_train,
            'step_val': step_val,
            'step_total': step_total,
        }
        return data

    def save(self, path: Union[str, Path]) -> None:
        """
        save model

        Parameters
        ----------
        path : str, Path
            The model home path
        """
        path = str(Path(path) / f'{self.model.name}.h5')
        self.model.save(path)

    def load(self, path: Union[str, Path]) -> None:
        """
        load model

        Parameters
        ----------
        path : str, Path
            The model home path
        """
        path = str(Path(path) / f'{self.model.name}.h5')
        self.model = tf.keras.models.load_model(path)


class FernTrainer(FernBaseTrainer):
    def __init__(
            self,
            model: FernModel,
            path_data: Union[str, Path],
            opt: str = 'adam',
            lr: float = 0.003,
            batch_size: int = 64,
            data_col: str = 'data',
            label_col: str = 'label'):
        """
        model trainer

        This can be used in one output and multi outputs

        Args:
            model: the model to be trained
            path_data: Points to training data in npz format
            opt: the name of optimizer
            lr: learning rate
            batch_size: batch size
            data_col: data column name
            label_col: label column name
        """
        self.model = model
        self.optimizer = tf.keras.optimizers.get({
            'class_name': opt,
            'config': {'learning_rate': lr}
        })

        self.metrics, self.early_stop_metric = self.setup_metrics()

        self.data = self.load_data(path_data, batch_size, data_col, label_col)

    def train(self, epochs: int = 1, early_stop: Optional[int] = None, mode: str = 'search') -> Tuple[float, int]:
        """
        train the model

        Args:
            epochs: epoch number
            early_stop:
                If None, no early stop will be used.
                If is number, after number times no update, will stop training early.
            mode:
                search: train model with data_train
                server: train model with data_train and data_val

        Returns:
            (best_score, best_epoch)
        """
        if mode == 'server':
            dataset = self.data['dataset_total']
            total = self.data['step_total']
        else:
            dataset = self.data['dataset_train']
            total = self.data['step_train']

        best_score = 0
        best_epoch = 0
        number = 0
        for epoch in range(epochs):
            # reset metrics before the next epoch
            for stage in self.metrics:
                for metric in self.metrics[stage]:
                    self.metrics[stage][metric].reset_states()
            number += 1

            for data, label in ProgressBar(dataset, desc='Train: ', total=total):
                self.train_step(data, label)

            for data, label in ProgressBar(self.data['dataset_val'], desc='Val: ', total=self.data['step_val']):
                self.val_step(data, label)

            stop_flag = False
            if early_stop is None:
                log = [f'Epoch: {epoch}']
            else:
                early_stop_result = self.early_stop_metric.result().numpy()
                if hasattr(early_stop_result, '__len__'):
                    score = early_stop_result[-1]
                else:
                    score = early_stop_result
                if score - best_score >= 0.0001:
                    best_score = score
                    best_epoch = epoch + 1
                    number = 0
                elif number == early_stop:
                    stop_flag = True
                log = [f"Epoch: {epoch}, Early Stop: {number}/{early_stop}({'Yes' if stop_flag else 'No'})"]

            for stage in self.metrics:
                log_line = [f'{stage}: ']
                for metric in self.metrics[stage]:
                    result = self.metrics[stage][metric].result()
                    log_line.append(f'{metric}: {result}')
                log_line = '\t'.join(log_line)
                log.append(log_line)
            log = '\n\t'.join(log)
            log += '\n'
            logger.info(log)

            if stop_flag:
                break
        return best_score, best_epoch

    @tf.function
    def train_step(
            self,
            data: Union[Dict[str, tf.Tensor], tf.Tensor],
            label: Union[Dict[str, tf.Tensor], tf.Tensor]
    ) -> None:
        """
        train one step

        Args:
            data: the input train data
            label: the output train labels
        """
        with tf.GradientTape() as tape:
            prediction = self.model(data)
            loss = self.loss(label, prediction)
        gradients = tape.gradient(loss, self.model.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.model.trainable_variables))
        acc = self.acc(label, prediction)

        self.metrics['train']['loss'].update_state(loss)
        self.metrics['train']['acc'].update_state(acc)

    @tf.function
    def val_step(
            self,
            data: Union[Dict[str, tf.Tensor], tf.Tensor],
            label: Union[Dict[str, tf.Tensor], tf.Tensor]
    ) -> None:
        """
        val one step

        Args:
            data: the input train data
            label: the output train labels
        """
        prediction = self.model(data)
        loss = self.loss(label, prediction)
        acc = self.acc(label, prediction)

        self.metrics['val']['loss'].update_state(loss)
        self.metrics['val']['acc'].update_state(acc)

    @staticmethod
    def setup_metrics() -> Tuple[Dict[str, Dict[str, Metric]], Metric]:
        """
        Initialize the evaluation metrics.
        If you define the metric yourself, please remember to edit your train_step and val_step

        Returns
        -------
        (dict[str, dict[str, Metric]], Metric)
            The dictionary with train and test, and each train or test with a dictionary of metric.
            And the other output of metric is for early stop.

        Examples
        --------
        ```python

        # Possible return format pseudo-code

        res = (
            {
                'train': {
                    'loss': MetricLoss,
                    'acc': MetricAcc,
                    'you_defined': ...},
                'val': {
                    'loss': MetricLoss,
                    'acc': MetricAcc,
                    'you_defined': ...},
            },
            EarlyStopMetric    # default is val_acc
        )

        ```
        """
        res = {
            'train': {
                'loss': tf.keras.metrics.Mean(name='train_loss'),
                'acc': tf.keras.metrics.Mean(name='train_acc')
            },
            'val': {
                'loss': tf.keras.metrics.Mean(name='val_loss'),
                'acc': tf.keras.metrics.Mean(name='val_acc'),
            }
        }
        return res, res['val']['acc']

    @staticmethod
    def loss(ys_desired: Union[Dict[str, tf.Tensor], tf.Tensor], ys_predicted: Union[Dict[str, tf.Tensor], tf.Tensor])\
            -> tf.Tensor:
        """
        loss function

        suitable for:
            1. single label (If it is 2 category case, then you also need to convert to one-hot format)
            2. single output (array), multiple output (dict[str, array])
        not suitable for:
            1. multiple label (you should change the loss function)

        Args:
            ys_desired: each with shape (None, m)
            ys_predicted: each with shape (None, m)

        Returns:
            with shape ()
        """
        assert type(ys_desired) is type(ys_predicted), 'Please make sure ys_desired and ys_predicted have same type'
        if isinstance(ys_desired, tf.Tensor):
            ys_predicted = {'label': ys_predicted}
            ys_desired = {'label': ys_desired}

        sum_res = 0
        for key in ys_desired:
            # if use multi label output, change to tf.losses.binary_crossentropy
            res = tf.losses.categorical_crossentropy(ys_desired[key], ys_predicted[key])
            sum_res += res
        sum_res = tf.reduce_mean(sum_res)
        return sum_res

    @staticmethod
    def acc(ys_desired: Union[Dict[str, tf.Tensor], tf.Tensor], ys_predicted: Union[Dict[str, tf.Tensor], tf.Tensor])\
            -> tf.Tensor:
        """
        accuracy function

        This function is for one output, multi outputs and for one-hot prediction.
        If you want to use multi-label prediction, you should rewrite this function.

        Args:
            ys_desired: each with shape (None, m)
            ys_predicted: each with shape (None, m)

        Returns:
            with shape (None, 1). When multi outputs, only while all outputs is True, the response will be True
        """
        assert type(ys_desired) is type(ys_predicted), 'Please make sure ys_desired and ys_predicted have same type'
        if isinstance(ys_desired, tf.Tensor):
            ys_predicted = {'label': ys_predicted}
            ys_desired = {'label': ys_desired}

        res = tf.constant(1)
        for key in ys_desired:
            # if use multi label output, you should change this code block
            y_true = tf.argmax(ys_desired[key], axis=-1)
            y_pred = tf.argmax(ys_predicted[key], axis=-1)
            res *= tf.cast(tf.equal(y_true, y_pred), tf.float32)
        return res


def compile_keras_model(
        model: tf.keras.Model,
        opt: str = 'adam',
        lr: float = 0.003,
        run_eagerly: Optional[int] = None,
        multi_classification: bool = True,
) -> tf.keras.Model:
    """
    编译keras模型

    Args:
        model: keras model
        opt: 优化器名字
        lr: 学习率
        run_eagerly: 如果是1, 用于动态图调试
        multi_classification: 如果是多标签的, 那么使用 binary_crossentropy; 否则使用

    Returns:
        编译后的模型
    """
    optimizer = tf.keras.optimizers.get({
        'class_name': opt,
        'config': {'learning_rate': lr}
    })

    if multi_classification:
        model.compile(
            optimizer=optimizer,
            loss=tf.keras.losses.binary_crossentropy,
            metrics=[
                tf.keras.metrics.binary_accuracy,
                tf.keras.metrics.AUC(),
            ],
            run_eagerly=run_eagerly,
        )
    else:
        model.compile(
            optimizer=optimizer,
            loss=tf.keras.losses.categorical_crossentropy,
            metrics=[
                tf.keras.metrics.categorical_accuracy,
                tf.keras.metrics.AUC(),
            ],
            run_eagerly=run_eagerly,
        )
    return model


class FernKerasModelTrainer(object):
    def __init__(self,
                 model: tf.keras.Model,
                 ckpt_home: str,
                 log_home: str,
                 train_ds: tf.data.Dataset,
                 val_ds: Optional[tf.data.Dataset] = None,
                 version: int = 1):
        self.model = model
        self.train_ds = train_ds
        self.val_ds = val_ds

        self.log_dir = Path(log_home) / f'{model.name}' / str(version)
        self.ckpt_dir = Path(ckpt_home) / f'{model.name}' / str(version) / 'weight.{epoch:02d}'

    def train(self, epochs: int = 1, steps_per_epoch: Optional[int] = None, verbose=1) -> tf.keras.callbacks.History:
        callbacks = [
            tf.keras.callbacks.TensorBoard(self.log_dir, update_freq=100, write_images=True, embeddings_freq=1),
            tf.keras.callbacks.ModelCheckpoint(self.ckpt_dir)
        ]
        kwargs = dict(
            x=self.train_ds,
            callbacks=callbacks,
            epochs=epochs,
            steps_per_epoch=steps_per_epoch,
            verbose=verbose
        )

        if self.val_ds:
            history = self.model.fit(validation_data=self.val_ds, **kwargs)
        else:
            history = self.model.fit(**kwargs)
        return history



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""data split"""
from typing import *

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split


def train_data_split(
        data: pd.DataFrame,
        test_size: Union[int, float],
        max_test_size: Optional[int] = None,
        random_state: Union[int, np.random.RandomState, None] = None
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    使用sklearn的方法分割数据集

    Args:
        data: 待分割数据
        test_size: 如果是int，那么test数据是具体数量；如果是float，那么test数据会是比例
        max_test_size: 如果test_size是float, 那么可以使用这个字段限制最大的测试集大小
        random_state: 随机种子

    Returns:
        (train_df, test_df)
    """
    if isinstance(test_size, float) and test_size <= 1:
        test_size = int(len(data) * test_size)
        if isinstance(max_test_size, int) and test_size > max_test_size:
            test_size = max_test_size
    train_df, test_df = train_test_split(data, test_size=test_size, random_state=random_state)
    return train_df, test_df



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""data tokenize"""
from typing import *
import re
from collections import Counter

import jieba
import pandas as pd
from sklearn.preprocessing import LabelBinarizer, MultiLabelBinarizer


SPECIAL_TOKEN = [
    '[pad]',    # 占位符
    '[cls]',    # 分类符
    '[sep]',    # 分割符号
    '[ukn]',    # 未知token
]


def str2word(string: str, zh_segmentation=True) -> List[str]:
    """
    分割字符串为词列表，支持的语言：中文、英文、数字、特殊符号（'-_.[]'），未支持的语言会被自动过滤

    操作顺序：
        1. 拆分中文和其他语言
        2. 除了中文、英文、数字、特殊符号（'-_.[]'）外，都作为分词分界线
        3. 针对中文：使用jieba分词
        4. 针对英文、数字：直接合成

    Args:
        string: the cleaned data
        zh_segmentation: 中文是否分词，默认使用结巴分词；否则按字分词

    Returns:
        cleaned word list
    """
    if zh_segmentation:
        string = re.sub(r'([\u4e00-\u9fa5]+)', r' \1 ', string)
        string = re.sub(r'(\[)(.+?)(])', r' \1\2\3 ', string)    # 确保特殊token的安全: xxx[yyy]zzz -> xxx [yyy] zzz

        string_list = re.split(r'[^a-zA-Z0-9\-_.\u4e00-\u9fa5\[\]]+', string)
        words = []
        for item in string_list:
            if not item:
                continue
            if re.match(r'[\u4e00-\u9fa5]', item):
                tmp = [word for word in jieba.cut(item) if word]
            else:
                tmp = [item]
            words.extend(tmp)
    else:
        string = re.sub(r'([\u4e00-\u9fa5])', r' \1 ', string)
        words = re.split(r'[^a-zA-Z0-9\-_.\u4e00-\u9fa5\[\]]+', string)
    return words


def generate_label_data(
        data: pd.DataFrame,
        label_col: str,
) -> Union[LabelBinarizer, MultiLabelBinarizer, Dict[str, Union[LabelBinarizer, MultiLabelBinarizer]]]:
    """
    利用sklearn工具加载标签字典，支持标签列的格式如下：
        1. 多任务：{'task1': [label1, label2], 'task2': label1}
        2. 单任务多标签： [label1, label2]
        3. 单任务单标签： label1

    Args:
        data: 带有标签列的data frame
        label_col: 标签列名字

    Returns:
        返回数据格式如下：
            1. 多任务：{'task1': MultiLabelBinarizer, 'task2': LabelBinarizer}
            2. 单任务多标签： MultiLabelBinarizer
            3. 单任务单标签： LabelBinarizer

    Raises:
        TypeError: 如果数据类型不正确，那么就会报错
    """
    label_example = data.loc[0, label_col]

    if isinstance(label_example, dict):
        # 多任务
        labels = {}

        def add_label(item):
            for key_ in item:
                if key_ in labels:
                    labels[key_].append(item[key_])
                else:
                    labels[key_] = [item[key_]]

        data[label_col].map(add_label)
        encoder = {}
        for key in labels:
            if isinstance(label_example[key], (list, tuple)):
                # 多标签
                encoder_ = MultiLabelBinarizer()
                encoder_.fit(labels[key])
                encoder[key] = encoder_
            elif isinstance(label_example[key], str):
                # 单标签
                encoder_ = LabelBinarizer()
                encoder_.fit(labels[key])
                encoder[key] = encoder_
    elif isinstance(label_example, (list, tuple)):
        # 单任务多标签
        labels = data[label_col].to_list()
        encoder = MultiLabelBinarizer()
        encoder.fit(labels)
    elif isinstance(label_example, str):
        # 单任务单标签
        labels = data[label_col].to_list()
        encoder = LabelBinarizer()
        encoder.fit(labels)
    else:
        raise TypeError('标签类型无法被支持')
    return encoder


def generate_word_library(data: pd.DataFrame, data_col: str, top: Optional[int] = None) -> Tuple[dict, dict]:
    """
    从数据集中加载词库，要求data_col列中的数据格式都是字符串

    Args:
        data: the data frame where the source data is stored.
        data_col: 数据加载源的列名，要求这个列里面都是字符串
        top: 按照词频排序，只提取前n个词。默认提取全部

    Returns:
        word2id, id2word字典

    Raises:
        ValueError: if no data provide, ValueError will be raised
    """
    if data is None:
        raise ValueError('No data is provided.')

    words = []
    words_append = words.append

    def _map_func(word_list):
        for word in word_list:
            word = word.strip().lower()
            if word and word[0] != '#':
                words_append(word)

    data[data_col].map(_map_func)
    counter = Counter(words)
    res = counter.most_common(n=top)

    res = [(token, 0) for token in SPECIAL_TOKEN] + res

    word2id = {item[0]: idx for idx, item in enumerate(res)}
    id2word = {idx: item[0] for idx, item in enumerate(res)}

    return word2id, id2word


def limit_token_length(tokens: list, /, n: int, strategy: str = 'tail') -> list:
    """
    限制token的长度.

    如果token列表长度超过n, 限制策略如下：
        1. head: 截取前n个token
        2. head+tail: 截取前n/3和后2n/3个token，并进行组合
        3. tail: 截取后n个token; 默认值

    注意：
        不建议在训练或者测试的时候移除token数量过多的样本，而是推荐截取的方式. 这样可以尽量训练和测试的样本格式的一致

    Args:
        tokens: 需要截取的token列表
        n: 最大列表长度
        strategy: 截取策略

    Returns:
        截取之后的token列表
    """
    if len(tokens) <= n:
        return tokens

    if strategy == 'head+tail':
        head = n//3
        res = tokens[:head] + tokens[head-n:]
    elif strategy == 'head':
        res = tokens[:n]
    else:
        res = tokens[-n:]
    return res



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""data utils"""
from typing import *
import re
import pickle
import pathlib

import pandas as pd

from fern.utils import check_path
from fern.data import FernDataFrame


def save_to_csv(data: Union[pd.DataFrame, FernDataFrame], path: Union[str, pathlib.Path]):
    """
    save data to path

    Args:
        data: 需要保存的data frame
        path: path where data save
    """
    check_path(path)
    if not data:
        raise ValueError('You should get source data before save')
    data = FernDataFrame(data)
    data.save(path)


def load_from_csv(path: Union[str, pathlib.Path], index_col: str = None, eval_col: Optional[List[str]] = None):
    """
    load data from path

    Args:
        path: path where data save
        index_col: 需要初始化的index 列
        eval_col: 需要恢复数据格式的数据列，读取的数据默认是string格式
    """
    data = pd.read_csv(path)
    if index_col:
        data = data.set_index(index_col)
    if isinstance(eval_col, list):
        for col in eval_col:
            data.loc[:, col] = data[col].map(eval)
    return data


def save_to_pickle(data: Any, path: Union[str, pathlib.Path]):
    """
    save data to path

    Args:
        data: 待保存的数据
        path: path where data save
    """
    check_path(path)
    if data is None:
        raise ValueError('You should get source data before save')
    with open(path, 'wb') as f:
        pickle.dump(data, f, protocol=4)


def load_from_pickle(path: Union[str, pathlib.Path]):
    """
    load data from path

    Args:
        path: path where data save
    """
    with open(path, 'rb') as f:
        data = pickle.load(f)
    return data


def read_words(words_path):
    """
    Read user words, stop words and word library from path

    Lines beginning with `#` or consisting entirely of white space characters will be ignored

    Parameters
    ----------
    words_path : str, Path, None
        words path

    Returns
    -------
    list[str]
        user word list and stop word list
    """

    def read(path):
        res = set()
        with open(path, mode='r', encoding='utf-8') as f:
            for line in f:
                line = line.strip().lower()
                if line and line[0] != '#':
                    res.add(line)
        res = list(res)
        return res

    if words_path is None or not pathlib.Path(words_path).exists():
        words = []
    else:
        words = read(words_path)
    return words


def read_regex_words(words_path):
    """
    Read words written through the regex

    Parameters
    ----------
    words_path : str, Path, None
        words path

    Returns
    -------
    list[re.Pattern]
        user word list and stop word list
    """
    words = read_words(words_path)
    word_reg = [re.compile(word) for word in words]
    return word_reg


def read_library_size(path):
    """
    read the length of the word/label library
    this will skip the space line automatically

    Parameters
    ----------
    path : str, pathlib.Path
        word library path

    Returns
    -------
    int
        length of the word library
    """
    words = read_words(path)
    return len(words)



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
from .data_class import *
from .data_utils import *

from .data_download import *
from .data_clean import *
from .data_tokenize import *
from .data_split import *
from .data_balance import *



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
""""model file"""
import logging
from tensorflow.keras import Model, layers


logger = logging.getLogger()


class FernModel(object):
    model: Model

    def __init__(self, output_shape, max_seq_len, library_len, initializer='he_normal'):
        """
        model builder

        Parameters
        ----------
        output_shape : dict[str, int], list[int], tuple[int]
            output shape without batch size
        max_seq_len : int
            the max input sequence length
        library_len : int
            the world library length
        initializer : str
            global initializer
        """
        self.output_shape = output_shape
        self.max_seq_len = max_seq_len
        self.library_len = library_len
        self.initializer = initializer
        self.name = self.__class__.__name__

        self.model = self.build()
        self.print_summary()

        self.compile = self.model.compile
        self.fit = self.model.fit

    def print_summary(self):
        """
        print summary of model
        """
        summary = []
        self.model.summary(print_fn=summary.append)
        summary = '\n'.join(summary)
        logger.info(f"\n{summary}")

    def build(self):
        """
        build model

        Returns
        -------
        Model
            built model
        """
        raise NotImplementedError

    def save(self, path):
        """
        save model

        Parameters
        ----------
        path : str, pathlib.Path
            The model file path
        """
        self.model.save(path)

    def load(self, path):
        """
        load model

        Parameters
        ----------
        path : str, pathlib.Path
            The model file path
        """
        self.model.load_weights(path)

    @property
    def predict(self):
        return self.model.predict

    def __call__(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    @property
    def trainable_variables(self):
        return self.model.trainable_variables


class TextCNN(FernModel):
    """
    References
    ----------
    Optimization of model Convolutional Neural Networks for Sentence Classification
     (https://arxiv.org/pdf/1408.5882.pdf)
    """
    def build(self):
        inp = layers.Input(shape=(self.max_seq_len,))
        x = layers.Embedding(self.library_len, 256, embeddings_initializer=self.initializer)(inp)
        x = layers.Conv1D(256, 5, padding='same', kernel_initializer=self.initializer)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Activation('relu')(x)

        x = layers.GlobalMaxPool1D()(x)

        x = layers.Dense(128, kernel_initializer=self.initializer)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Activation('relu')(x)

        ys = []
        for key in self.output_shape:
            y = layers.Dense(self.output_shape[key],
                             kernel_initializer=self.initializer,
                             activation='softmax',
                             name=key)(x)
            ys.append(y)
        model = Model(inputs=inp, outputs=ys, name=self.name)    # if len(ys) == 1, than ys = ys[0]
        return model



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""pipelines tools"""
from typing import *

import numpy as np
import pandas as pd
from sqlalchemy import create_engine

from fern.data import FernDataFrame


class FernDownloader(object):
    def read_sql(self, url: str, sql: str, index: Optional[str] = None, drop: bool = False) -> FernDataFrame:
        """
        通过sql, 从数据库读取数据

        Args:
            url: 用于连接数据库, 例如: 'mysql+pymysql://user:passwd@hostname/dbname?charset=utf8mb4'
            sql: 读取数据库的sql语句
            index: 如果提供, 会根据这个字段, 创建返回的dataframe的index列
            drop: 默认不移除空白行

        Returns:
            从数据库中下载的原始数据
        """
        raise NotImplementedError


class FernCleaner(object):
    """
    data cleaner
    """
    def clean(self, data: Union[pd.DataFrame, FernDataFrame]) -> FernDataFrame:
        """
        clean data main entry, 如果清洗数据过程中需要用到多进程，那么需要保证被调用函数是静态的

        一般流程：
            1. 清洗原始数据
            2. 清洗label数据

        Args:
            data: source data frame

        Returns:
            cleaned data with columns, data_col and label_col
        """
        raise NotImplementedError

    def clean_data(self, row: Union[pd.Series, dict]) -> Any:
        """
        convert input data function for pandas apply function

        due to the limitation of parallel process, the function need to be static method

        一般流程：
            1. 关键数据截取
            2. 特殊字符清理
            3. 停用词移除(不在此处操作, 这个在tokenize中操作)

        Args:
            row: the input data row

        Returns:
            清理好的文本数据。通常不同文本段这件拼接使用空格，这样在下一步操作的时候，这个空格就会被直接移除
        """
        raise NotImplementedError

    def clean_label(self, row: Union[pd.Series, dict]) -> Any:
        """
        clean label function for pandas apply function

        一般流程：
            1. 只做标签格式化操作，通常是大小写等简单操作

        注意事项：
            1. 如果不是必要可以不要做标签清理
            2. 如果是单任务单标签可以整理成 label1 的数据格式
            3. 如果是单任务多标签形式，那么可以整理成 [label1, label2]的数据格式
            4. 如果是多任务形式，那么可以整理成 {'task1': [label1, label2], 'task2': [label1, label2]}的数据格式

        Args:
            row: find source label data

        Returns:
            cleaned label
        """
        raise NotImplementedError


class FernTokenizer(object):
    """
    转换清洗完的数据为token列表格式，不涉及到pad和生成array的操作
    """
    def tokenize(self, data: Union[pd.DataFrame, FernDataFrame]) -> Union[pd.DataFrame, FernDataFrame]:
        """
        转化一个data col 和label col对应的数据为token格式，但是不做padding的操作

        一般操作流程：
            1. token data
            2. token label
            3. 保存数据

        Args:
            data: the data frame where the cleaned data is stored

        Returns:
            The transformed word id sequence
        """
        raise NotImplementedError

    def tokenize_data(self, string: str) -> np.ndarray:
        """
        转化string文本为token列表格式

        一般操作步骤：
            1. 文本分词为词列表
            2. 使用停用词库移除停用词
            3. 词库中未出现的词, 转化为ukn token
            2. 如果必要，可以添加特殊token到这个列表中
            2. 使用word2id转化为token列表

        Args:
            string: 待转化文本

        Returns:
            The transformed word id sequence
        """
        raise NotImplementedError

    def tokenize_label(self, label: Union[dict, list, tuple, str]) -> Union[dict, np.ndarray]:
        """
        转化label为id array格式

        Args:
            label: to be transform source label data

        Returns:
            如果是多任务，那么label格式是字典，此时返回数据也需要是字典；array对象
        """
        raise NotImplementedError


class FernSplitter(object):
    """split data into train data and val data"""
    def __init__(self, rate_val: float, random_state: Optional[int] = None):
        self.rate_val = rate_val
        self.random_state = random_state

    def split(self, data: Union[pd.DataFrame, FernDataFrame]) -> Union[pd.DataFrame, FernDataFrame]:
        """
        split function to split data

        一般流程:
            1. 分割数据集为训练和评估数据集 (一般测试数据集会在更早之前被分割出去)
            2. 保存数据集

        Args:
            data: 待分割数据集
        """
        raise NotImplementedError


class FernBalance(object):
    """
    Class for sample balance
    """
    def balance(self, data: Union[pd.DataFrame, FernDataFrame]) -> Union[pd.DataFrame, FernDataFrame]:
        """
        根据类别数量，平衡样本数量

        一般流程：
            1. 制作平衡好的数据集
            2. 保存数据

        Args:
            data: 需要平衡的数据集
        """
        raise NotImplementedError



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""common function"""
import logging
import pathlib
import subprocess
from typing import *

import tensorflow as tf
from tqdm import tqdm
import yaml


logger = logging.getLogger('Fern')


def check_path(path):
    """
    check if path exits. If not exit, the path.parent will be created.

    Parameters
    ----------
    path : str, Path
        path to be check
    """
    path = pathlib.Path(path).parent
    if not path.exists():
        path.mkdir(parents=True)


def get_available_gpu(min_memory=0):
    """
    find the gpu of which free memory is largest

    References
    ----------
    refer to link: https://stackoverflow.com/a/59571639

    Parameters
    ----------
    min_memory : int
        Minimum allowable memory in MB

    Returns
    -------
    tuple[int, int] or None
        - If there is gpu available, return (gpu index, free memory)
        - Else return (None, 0)
    """
    # output: b'memory.free [MiB]\n48600 MiB\n48274 MiB\n'
    command = "nvidia-smi --query-gpu=memory.free --format=csv"
    try:
        memory_free_info = subprocess.check_output(command.split())
    except FileNotFoundError:
        return None, 0
    memory_info = memory_free_info.decode('ascii').split('\n')[1:-1]
    res = []
    for i, memory in enumerate(memory_info):
        memory = int(memory.split()[0])
        if memory > min_memory:
            res.append((i, memory))
    res = sorted(res, key=lambda item: item[1], reverse=True)
    if res:
        return res[0]
    else:
        return None, 0


def set_gpu(index: Optional[int] = None, growth: bool = True):
    """
    set which GPU to use

    Args:
        index: the gpu index
        growth: whether to limit gpu memory growth
    """
    gpus = tf.config.experimental.list_physical_devices('GPU')
    if growth:
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)

    if gpus and isinstance(index, int):
        try:
            tf.config.experimental.set_visible_devices(gpus[index], 'GPU')
        except RuntimeError as e:
            # Memory growth must be set before GPUs have been initialized
            logger.info(e)
        except IndexError as e:
            # there is no such a gpu found
            logger.info(e)


class ProgressBar(tqdm):
    def __init__(self, *arg, **kwargs):
        super().__init__(ascii='->', leave=False, *arg, **kwargs)


def read_config(path: str):
    """读取yaml配置文件"""
    with open(path, 'r') as f:
        data = yaml.safe_load(f)
    return data



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
from typing import *

import pandas as pd
from sqlalchemy import create_engine

from fern.data import FernDataFrame


def read_sql(url: str, sql: str, index: Optional[str] = None, drop: bool = False) -> FernDataFrame:
    """
    通过sql, 从数据库读取数据

    Args:
        url: 用于连接数据库, 例如: 'mysql+pymysql://user:passwd@hostname/dbname?charset=utf8mb4'
        sql: 读取数据库的sql语句
        index: 如果提供, 会根据这个字段, 创建返回的dataframe的index列
        drop: 默认不移除空白行

    Returns:
        从数据库中下载的原始数据
    """
    engine = create_engine(url=url)

    df = pd.read_sql(sql, engine)
    df = FernDataFrame(df)

    if drop:
        df = df.dropna()
    if index is None:
        data = df
    else:
        data = df.set_index(index)
    return data



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""layer model"""
import typing
import re
import pathlib

import tensorflow as tf
import tensorflow_text as text
from tensorflow.keras import layers


class Conv1DPassMask(layers.Conv1D):
    """to pass mask though conv1d"""
    def compute_mask(self, inputs, mask=None):
        return mask


class FlattenPassMask(layers.Flatten):
    """传递mask"""
    def compute_mask(self, inputs, mask=None):
        mask = tf.expand_dims(mask, axis=-1)
        mask = tf.concat([mask] * inputs.shape[-1], axis=-1)
        mask = tf.reshape(mask, (-1, mask.shape[-2] * mask.shape[-1]))
        return mask


class DenseWithMask(layers.Dense):
    def call(self, inputs, mask=None):
        if mask is not None:
            mask = tf.cast(mask, dtype=tf.keras.backend.floatx())
            inputs *= mask
            inputs /= tf.reduce_sum(mask, axis=-1, keepdims=True)
        res = super().call(inputs)
        return res


class AttentionLayer(tf.keras.layers.Layer):
    """to replace global max pool"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.weight = None  # type: typing.Optional[tf.keras.layers.Layer]

    def build(self, input_shape):
        """input_shape = (None, 25, 256)"""
        assert len(input_shape) == 3, str(input_shape)
        self.weight = self.add_weight(name='att_weight',
                                      shape=(input_shape[-2], input_shape[-2]),
                                      initializer='uniform',
                                      trainable=True)
        super().build(input_shape)

    def call(self, inputs):
        """inputs.shape = (None, 25, 256)"""
        w = tf.transpose(inputs, perm=(0, 2, 1))
        w = tf.linalg.matmul(w, self.weight)
        w = tf.math.softmax(w)
        outputs = tf.linalg.matmul(w, inputs)
        outputs = tf.math.reduce_sum(outputs, axis=-2)
        return outputs

    @staticmethod
    def compute_output_shape(input_shape):
        res = input_shape[:-2] + input_shape[-1]
        return res

    @staticmethod
    def compute_mask(_, mask=None):
        return mask


class ScaledDotProductAttention(layers.Layer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def build(self, input_shape):
        """input_shape = (None, 25, 256)"""
        super().build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        """inputs.shape = (None, 25, 256)"""
        q = k = v = inputs
        feature_dim = inputs.shape[-1]

        kt = tf.transpose(k, perm=(0, 2, 1))
        res = tf.linalg.matmul(q, kt)
        res = res / (feature_dim ** 0.5)
        if mask is not None:
            mask = tf.expand_dims(mask, axis=-1)
            mask = tf.cast(mask, dtype=tf.keras.backend.floatx())
            res *= mask
        res = tf.math.softmax(res)
        res = tf.linalg.matmul(res, v)
        return res

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_mask(self, inputs, mask=None):
        return mask


class TransposeLayer(layers.Layer):
    """Transpose Layer"""
    def call(self, inputs, mask=None, **kwargs):
        y = tf.transpose(inputs, perm=(0, 2, 1))
        return y

    def compute_output_shape(self, input_shape):
        res = [input_shape[0], input_shape[2], input_shape[1]]
        return res


class BertTokenizer(tf.Module):
    """
    根据vocab字典转化token
    """
    def __init__(self, vocab_path, suffix_indicator='##', name='bert-tokenizer'):
        super().__init__(name=name)
        self.tokenizer = text.BertTokenizer(vocab_path, suffix_indicator=suffix_indicator, lower_case=True)
        self._suffix_indicator = suffix_indicator
        self._vocab_path = tf.saved_model.Asset(vocab_path)

        vocab = pathlib.Path(vocab_path).read_text().splitlines()
        self.vocab = tf.Variable(vocab)

    def init_signature(self):
        # 看似没有效果, 这样就可以实现模型的自动初始化了(可以使用tf save_model导出)
        self.tokenize.get_concrete_function(
            tf.TensorSpec(shape=[None], dtype=tf.string))

        # Include `detokenize` and `lookup` signatures for:
        #   * `Tensors` with shapes [tokens] and [batch, tokens]
        #   * `RaggedTensors` with shape [batch, tokens]
        self.detokenize.get_concrete_function(
            tf.TensorSpec(shape=[None, None], dtype=tf.int64))
        self.detokenize.get_concrete_function(
            tf.RaggedTensorSpec(shape=[None, None], dtype=tf.int64))

        self.lookup.get_concrete_function(
            tf.TensorSpec(shape=[None, None], dtype=tf.int64))
        self.lookup.get_concrete_function(
            tf.RaggedTensorSpec(shape=[None, None], dtype=tf.int64))

        # These `get_*` methods take no arguments
        self.get_vocab_size.get_concrete_function()
        self.get_vocab_path.get_concrete_function()
        self.get_reserved_tokens.get_concrete_function()

    @tf.function
    def tokenize(self, strings):
        enc = self.tokenizer.tokenize(strings)
        # Merge the `word` and `word-piece` axes.
        enc = enc.merge_dims(-2, -1)
        return enc

    @tf.function
    def detokenize(self, tokenized):
        txt_tokens = tf.gather(self.vocab, tokenized)
        # Join with spaces.
        words = tf.strings.reduce_join(txt_tokens, separator=' ', axis=-1)
        words = tf.strings.regex_replace(words, re.escape(' ' + self._suffix_indicator))
        return words

    @tf.function
    def lookup(self, token_ids):
        return tf.gather(self.vocab, token_ids)

    @tf.function
    def get_vocab_size(self):
        return tf.shape(self.vocab)[0]

    @tf.function
    def get_vocab_path(self):
        return self._vocab_path

    @tf.function
    def get_reserved_tokens(self):
        mask = tf.strings.regex_full_match(self.vocab, r'\[[A-Z]+\]')
        res = tf.boolean_mask(self.vocab, mask)
        return res



# Fern
#
# Author: Jason, Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""loss"""
import tensorflow as tf
from tensorflow.keras.losses import Loss
from tensorflow.keras.backend import epsilon
from tensorflow import math


class BinaryFocalLoss(Loss):
    """
    针对多标签分类/2分类的 focal loss
    """
    def __init__(self, gamma: int = 2, **kwargs):
        super().__init__(**kwargs)
        self.gamma = gamma

    def call(self, y_true, y_pred):
        y_pred = tf.clip_by_value(y_pred, clip_value_min=epsilon(), clip_value_max=1-epsilon())  # 限制范围[ep, 1-ep]
        y_ = y_true * y_pred + (1.0 - y_true) * (1.0 - y_pred)  # 如果在这里使用
        loss = -((1 - y_) ** self.gamma) * math.log(y_)
        loss = math.reduce_sum(loss, axis=-1)
        return loss


class CategoricalFocalLoss(Loss):
    """
    针对多分类的 focal loss
    """
    def __init__(self, gamma: int = 2, **kwargs):
        super().__init__(**kwargs)
        self.gamma = gamma

    def call(self, y_true, y_pred):
        y_pred = tf.clip_by_value(y_pred, clip_value_min=epsilon(), clip_value_max=1-epsilon())  # 限制范围[ep, 1-ep]
        loss = -y_true * ((1 - y_pred)**self.gamma) * math.log(y_pred)
        loss = math.reduce_sum(loss, axis=-1)
        return loss



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""data cleaning"""


def full_width_to_half_width(string: str) -> str:
    """
    字符串全角转半角，例如：`１２３４５６７８９０ -> 1234567890`

    Args:
        string: 待转换字符串

    Returns:
        半角字符串
    """
    table = {ord(f): ord(t) for f, t in zip(
        u'＊・，。！？【】（）％＃＠＆１２３４５６７８９０、；—：■　ａｂｃｄｅｆｇｈｉｊｋｌｍｎｏｐｑｒｓｔｕｖｗｘｙｚＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ',
        u'*·,.!?[]()%#@&1234567890.;-:￭ abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')}

    new_string = ''
    for char in string:
        if char_ord := ord(char) in table:
            char_ord = table[char_ord]
            char = chr(char_ord)
        new_string += char
    return new_string


def string_to_camel(string: str) -> str:
    """
    把下划线字符串转为驼峰模式
    """
    return ''.join([_string.capitalize() for _string in string.split('_')])



# Fern
#
# Author: Jason, Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""metric"""
import tensorflow as tf
from tensorflow.keras.metrics import Metric


class BinaryCategoricalAccuracy(Metric):
    """
    To calculate the accuracy of the m outputs, when the output is multi-output matrix with shape (None, m, n)

    Parameters
    ----------
    m : int
        the number of output
    """
    def __init__(self, m, name='binary_categorical_accuracy', **kwargs):
        super().__init__(name=name, **kwargs)
        self.count = self.add_weight('count', initializer='zeros')
        self.values = self.add_weight('values', shape=(m, ), initializer='zeros')
        self.m = m

    def update_state(self, y_true, y_pred):
        """
        update state

        Parameters
        ----------
        y_true : tf.Tensor
            with shape (None, m, n)
        y_pred : tf.Tensor
            with shape (None, m, n)
        """
        y_true = tf.argmax(y_true, axis=-1)
        y_pred = tf.argmax(y_pred, axis=-1)

        value = tf.equal(y_true, y_pred)
        value = tf.cast(value, dtype=self.dtype)
        value = tf.reduce_sum(value, axis=0)

        self.values.assign_add(value)
        self.count.assign_add(y_true.shape[0])
    
    def reset_states(self):
        self.count.assign(0)
        self.values.assign([0] * self.m)

    def result(self):
        """
        return mean result
        """
        return self.values / self.count



# Fern
#
# Author: Jason Lin
# Email: jason.m.lin@outlook.com
#
# =============================================================================
"""setup file for packaging"""
import re
import os
import setuptools


REF = os.getenv('GITHUB_REF', '')   # refs/tags/1.1.1.dev5
VERSION = REF.split('/')[-1]
# check version, 1.1.1rc2.post1.dev1
if not VERSION or not re.match(r'^\d+\.\d+\.\d+(?:(?:a|b|rc)\d+)?(?:\.post\d+)?(?:\.dev\d+)?$', VERSION):
    raise ValueError(f'Version check failed: {VERSION}')

with open('README.md', 'r') as f:
    LONG_DESCRIPTION = f.read()

with open('requirements.txt', 'r') as f:
    INSTALL_REQUIRES = [line.strip() for line in f if line.strip()]

setuptools.setup(
    name='Fern2',
    version=VERSION,
    author='Jason, Lin',
    author_email='jason.m.lin@outlook.com',
    license='Apache 2.0',
    description='NLP text processing toolkit for Deep Learning',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/Jasonsey/Fern',
    packages=setuptools.find_packages(),
    install_requires=INSTALL_REQUIRES,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: MacOS',
        'Operating System :: Microsoft',
        'Operating System :: POSIX :: Linux',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Natural Language :: Chinese (Simplified)',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.6',
)



# Fern
# 
# @Author: Lin, Max
# @Email : max_lin1@dell.com
# @Time  : 2021/9/27 12:19 下午
#
# =============================================================================
"""tensor_ops.py"""
from typing import Union, Callable
import tensorflow as tf


@tf.function
def map_flat_values(map_fn: Callable, tensor: Union[tf.Tensor, tf.RaggedTensor]):
    """
    对于每个tensor, 实现flatten再传值, 然后保持前后shape一致

    Args:
        map_fn: map函数
        tensor: 输入的tensor

    Returns:
        处理好的tensor
    """
    if isinstance(tensor, tf.RaggedTensor):
        output = tf.ragged.map_flat_values(map_fn, tensor)
    else:
        # Tensor
        before_shape = tensor.shape.as_list()

        tensor = tf.reshape(tensor, (-1,))
        output = map_fn(tensor)

        after_shape = before_shape + output.shape.as_list()[1:]
        if after_shape[-1] == 1:
            # 如果最后一个维度是1, 那么会清理它
            after_shape.pop(-1)
        output = tf.reshape(output, after_shape)
    return output
